import csv
import json
import logging
from io import BytesIO

from pandas import DataFrame, ExcelWriter

from django.contrib import messages
from django.contrib.postgres.aggregates import StringAgg
from django.db.models import Max, OuterRef, Q, Subquery
from django.http import Http404, HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, View

from core.models import ActivityLogChange, ActivityLogEntry
from core.models.assemblies import Assembly, AssemblyMember
from core.models.conference import ConferenceExportCache
from core.models.users import UserCommunicationChannel
from core.views.list_views import AlphabeticalPaginatorView, FilteredListView

from backoffice.forms import AssemblyTeamMessageForm
from backoffice.templatetags.c3assemblies import get_language_item

from .mixins import AssemblyMixin, ConferenceLoginRequiredMixin

logger = logging.getLogger(__name__)


class AssemblyTeamMixin(ConferenceLoginRequiredMixin):
    require_conference = True
    permission_required = ['core.assembly_team']
    active_page = 'assemblies'
    base_view_name = 'backoffice:assemblies'
    list_view_name = 'backoffice:assemblieslist'
    sidebar_caption = _('nav_assemblies')
    status_field = 'state'

    MODES = {
        'all': (('XMode', Q(state__exact=Assembly.State.HIDDEN)), _('nav_assemblies_all')),
        'accepted': (('QMode', Q(state__in=Assembly.PUBLIC_STATES)), _('nav_assemblies_accepted')),
        'pending': (('QMode', Q(state__in=[Assembly.State.REGISTERED])), _('nav_assemblies_pending')),
        'planned': (('QMode', Q(state__in=[Assembly.State.PLANNED])), _('nav_assemblies_planned')),
        'rejected': (('QMode', Q(state__in=[Assembly.State.REJECTED])), _('nav_assemblies_rejected')),
        'hidden': (('QMode', Q(state__in=[Assembly.State.HIDDEN])), _('nav_assemblies_hidden')),
    }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['active_page'] = self.active_page
        self.request.session['assembly_back'] = {'link': self.request.get_full_path(), 'title': str(self.sidebar_caption)}

        assemblies = []
        lists = []
        context['sidebar'] = {
            'title': self.sidebar_caption,
            'items': [
                {
                    'caption': _('Assemblies'),
                    'children': assemblies,
                    'expanded': True,
                },
                {
                    'caption': _('lists'),
                    'children': lists,
                    'expanded': True,
                },
            ],
        }

        for m, (q, t) in self.MODES.items():
            assemblies.append(
                {
                    'mode': m,
                    'caption': t,
                    'count': self.conference.assemblies.filter(q[1]).count(),
                    'link': reverse(self.base_view_name) + '?mode=' + m,
                }
            )

        lists.append(
            {
                'caption': 'slug, name, friends & WA',
                'link': reverse(self.list_view_name, kwargs={'variant': 'slugname'}) + '?mode=accepted',
                'variant': 'slugname',
            }
        )

        lists.append(
            {
                'caption': 'contacts',
                'link': reverse(self.list_view_name, kwargs={'variant': 'assemblycontacts'}) + '?mode=accepted',
                'variant': 'assemblycontacts',
            }
        )

        lists.append(
            {
                'caption': 'contact mails',
                'link': reverse(self.list_view_name, kwargs={'variant': 'contactsmail'}) + '?mode=accepted',
                'variant': 'contactsmail',
            }
        )

        if self.conference.additional_fields_schema is not None:
            lists.append(
                {
                    'caption': 'registration',
                    'link': reverse(self.list_view_name, kwargs={'variant': 'registration'}) + '?mode=accepted',
                    'variant': 'registration',
                }
            )

        return context


class AssembliesListMixin(AssemblyTeamMixin, FilteredListView):
    default_assemblies_mode = 'all'
    filters = {
        'tag': ('tags__tag__slug', 'list'),
    }

    def get_queryset(self, **kwargs):
        # get the mode
        mode = (self.request.POST if self.request.method == 'POST' else self.request.GET).get('mode', self.default_assemblies_mode)
        query_filters = {}

        if mode in self.MODES:
            query_filters[self.MODES[mode][0][0]] = self.MODES[mode][0][1]
        else:
            messages.warning(self.request, f'unknown mode "{mode}", using "all"')
            mode = 'all'
            query_filters[self.MODES[mode][0][0]] = self.MODES[mode][0][1]
        self.assemblies_mode = mode

        qs = super().get_queryset(
            Assembly.objects.associated_with_user(conference=self.conference, user=self.request.user, staff_can_see=True), query_filters=query_filters
        )

        return qs


def build_nav_from_assembly(assembly):
    def _build_nav_from_assembly(assmbly: Assembly, way_up: bool):
        # for cluster members, go way up to the top and trickle down
        if way_up and assmbly.parent:
            return _build_nav_from_assembly(assmbly.parent, way_up=True)

        me = {
            'caption': assmbly.slug,
            'link': reverse('backoffice:assemblyteam-detail', kwargs={'pk': assmbly.pk}),
            'active': assembly.id == assmbly.id,
        }

        if assmbly.is_cluster:
            me['children'] = [_build_nav_from_assembly(a, way_up=False) for a in assmbly.children.order_by('slug').all()]
            me['expanded'] = True

        return [me] if way_up else me

    return _build_nav_from_assembly(assembly, way_up=True)


class SingleAssemblyTeamMixin(AssemblyMixin, AssemblyTeamMixin):
    assembly_management = True
    model = Assembly

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['sidebar'] = {
            'title': _('nav_assemblies'),
            'title_link': reverse('backoffice:assemblies'),
            'items': build_nav_from_assembly(self.assembly),
        }
        return ctx


class AssemblyView(SingleAssemblyTeamMixin, DetailView):
    template_name = 'backoffice/assemblyteam_assembly_detail.html'

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['latest_note'] = self.assembly.logentries.exclude(comment=None).order_by('-timestamp').first()
        return ctx

    def post(self, *args, **kwargs):
        comment = self.request.POST.get('comment', '').strip()
        if comment:
            self.assembly.log_activity(
                user=self.request.user,
                kind=ActivityLogEntry.Kind.STAFF,
                comment=comment,
            )
            messages.success(self.request, _('AssemblyTeam__add_comment_success'))
        return redirect('backoffice:assemblyteam-detail', self.assembly.id)


class AssembliesView(AssembliesListMixin, AlphabeticalPaginatorView):
    template_name = 'backoffice/assembly_list.html'

    def get_queryset(self, **kwargs):
        qs = (
            super()
            .get_queryset(**kwargs)
            .annotate(
                last_activity_log_staff=Max(
                    'activity_log_entries__timestamp', filter=Q(activity_log_entries__kind__in=[ActivityLogEntry.Kind.STAFF, ActivityLogEntry.Kind.MESSAGE])
                ),
                last_activity_log_assembly=Max('activity_log_entries__timestamp', filter=Q(activity_log_entries__kind=ActivityLogEntry.Kind.ENTITY)),
            )
        )
        return qs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # context['filter_tag'] = self.selected_tag

        context['mode'] = self.assemblies_mode
        context['mode_display'] = self.MODES[self.assemblies_mode][1]

        # activate current sidebar item
        for sidebar_item in context['sidebar']['items'][0]['children']:
            if sidebar_item['mode'] == self.assemblies_mode:
                sidebar_item['active'] = True
                break

        for obj in context['object_list']:
            obj.user_can_edit = property(lambda x: x.user_can_manage(self.request.user, staff_can_manage=True))

        return context


class AssembliesListsView(AssembliesListMixin):
    DEFAULT_DELIMITER_CHAR = ';'
    DEFAULT_QUOTE_CHAR = '"'
    DELIMITER_CHARS = ',;|\t=*/'
    QUOTE_CHARS = '"\'|*'

    def _get_request_char(self, param, valid_chars, default):
        """
        Reads the named GET/POST parameter and checks that it's a single character contained in valid_chars.
        Otherwise return default (which does not need to abide to these rules).
        """
        value = (self.request.POST if self.request.method == 'POST' else self.request.GET).get(param)
        if value is None:
            return default

        if len(value) != 1:
            if value == 'TAB':
                value = '\t'
            else:
                return default

        if value not in valid_chars:
            return default

        return value

    def prepare(self, *args, **kwargs):
        variant = kwargs.get('variant')
        variant_name = None
        variant_fields = None

        if variant == 'slugname':
            qs, variant_name, variant_fields = self.prepare_slug_name()
        elif variant == 'contactsmail':
            qs, variant_name, variant_fields = self.prepare_contacts_mail()
        elif variant == 'assemblycontacts':
            qs, variant_name, variant_fields = self.prepare_assembly_contacts()
        elif variant == 'registration':
            qs, variant_name, variant_fields = self.prepare_registration()
        else:
            return (None, HttpResponse('Unknown list variant.', status=404), None, None)

        return (qs, variant, variant_name, variant_fields)

    def prepare_assembly_contacts(self):
        # all assemblies with their associated contacts
        user_mails = Subquery(
            UserCommunicationChannel.objects.filter(
                user_id=OuterRef('member_id'),
                is_primary=True,
                is_verified=True,
                channel=UserCommunicationChannel.Channel.MAIL,
            ).values('address')
        )
        qs = (
            AssemblyMember.objects.filter(assembly__in=self.get_queryset(), can_manage_assembly=True)
            .annotate(mail=StringAgg(user_mails, '; '))
            .values_list('assembly__slug', 'is_representative', 'mail')
            .order_by('assembly__slug', 'mail')
        )

        variant_name = 'contact emails (assembly managers)'
        variant_fields = [_('Assembly__slug'), _('AssemblyMember__is_representative'), _('UserCommunicationChannel__address')]
        return (qs, variant_name, variant_fields)

    def prepare_contacts_mail(self):
        """
        Return all assembly contacts' email addresses with duplicates removed
        """
        qs = (
            UserCommunicationChannel.objects.filter(
                channel=UserCommunicationChannel.Channel.MAIL,
                is_primary=True,
                is_verified=True,
                user_id__in=AssemblyMember.objects.filter(assembly__in=self.get_queryset(), can_manage_assembly=True).values('member_id'),
            )
            .values('address')
            .distinct()
            .order_by('address')
        )

        variant_name = 'assembly contacts emails'
        variant_fields = [_('UserCommunicationChannel__address')]
        return (qs, variant_name, variant_fields)

    def prepare_registration(self):
        # assemblies with the (additional) registration data as columns
        qs_fields = ['slug', 'name', self.status_field, 'hierarchy', 'parent__slug', 'assembly_links_agg', 'description', 'registration_details']
        qs = (
            self.get_queryset()
            .prefetch_related('assembly_links')
            .annotate(
                assembly_links_agg=StringAgg('assembly_links__b__slug', delimiter=', ', distinct=True),
            )
        )
        qs = qs.only(*([*qs_fields, 'registration_data']))  # query qs_fields + additional data

        variant_name = 'registration details'
        variant_fields = [_(f'Assembly__{fld}') for fld in qs_fields]  # noqa: INT001

        for field in self.conference.additional_fields_schema or []:
            f_key = field['key']
            f_caption = get_language_item(field['caption'])
            f_lookup = f'registration_data__{f_key}'
            variant_fields.append(f_caption)
            qs_fields.append(f_lookup)
        qs = qs.values_list(*qs_fields).order_by('slug')
        return (qs, variant_name, variant_fields)

    def prepare_slug_name(self, *args, **kwargs):
        # all assemblies with slug + name + related assemblies
        qs = tuple(
            (a.slug, a.name, a.parent, ', '.join(link.b.slug for link in a.assembly_links.all()))
            for a in self.get_queryset().prefetch_related('assembly_links')
        )
        variant_name = 'slug, name, parent, link'
        variant_fields = [_('Assembly__slug'), _('Assembly__name'), _('Assembly__parent'), _('assembly_links')]
        return (qs, variant_name, variant_fields)

    def get(self, *args, **kwargs):
        qs, variant, variant_name, variant_fields = self.prepare(*args, **kwargs)
        self.object_list = qs
        if qs is None:
            assert isinstance(variant, HttpResponse)
            return variant

        data = list(qs)

        ctx = self.get_context_data()
        ctx['mode'] = self.assemblies_mode
        ctx['modes'] = list(self.MODES)
        ctx['variant'] = variant
        ctx['list_count'] = len(data)
        ctx['list_name'] = variant_name
        ctx['fields'] = variant_fields
        ctx['data'] = data

        # activate current sidebar item
        for sidebar_item in ctx['sidebar']['items'][1]['children']:
            if sidebar_item['variant'] == variant:
                sidebar_item['active'] = True
                break

        ctx['download_available'] = len(data) > 0
        ctx['download_default_delimiter'] = self.DEFAULT_DELIMITER_CHAR
        ctx['download_default_quotechar'] = self.DEFAULT_QUOTE_CHAR
        ctx['download_delimiters'] = [c for c in self.DELIMITER_CHARS if c != '\t']
        ctx['download_quotechars'] = list(self.QUOTE_CHARS)

        return render(self.request, 'backoffice/assembly_list_export.html', context=ctx)

    def post(self, *args, **kwargs):
        qs, variant, variant_name, variant_fields = self.prepare(*args, **kwargs)
        if qs is None:
            assert variant is HttpResponse
            return variant

        df = DataFrame.from_records(list(qs))
        date_columns = df.select_dtypes(include=['datetime64[ns, UTC]']).columns
        for date_column in date_columns:
            df[date_column] = df[date_column].dt.date

        export_type = self.request.POST.get('type')

        with BytesIO() as b:
            if export_type == 'csv':
                df.to_csv(
                    b,
                    sep=self._get_request_char('delimiter', self.DELIMITER_CHARS, self.DEFAULT_DELIMITER_CHAR),
                    quotechar=self._get_request_char('quote', self.QUOTE_CHARS, self.DEFAULT_QUOTE_CHAR),
                    quoting=csv.QUOTE_MINIMAL if self.request.POST.get('quoteall') is None else csv.QUOTE_ALL,
                    header=variant_fields if self.request.POST.get('headers') else False,
                )
                content_type = 'text/csv'
            else:
                with ExcelWriter(b, engine='openpyxl' if export_type == 'xlsx' else 'odf') as writer:
                    df.to_excel(
                        writer,
                        sheet_name='data',
                        index=False,
                        header=variant_fields if self.request.POST.get('headers') else False,
                    )
                content_type = (
                    ('application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',)
                    if export_type == 'xlsx'
                    else ('application/vnd.oasis.opendocument.spreadsheet',)
                )
            filename = f'{variant_name}.{export_type}'
            res = HttpResponse(
                b.getvalue(),  # Gives the Byte string of the Byte Buffer object
                content_type=content_type,
            )
            res['Content-Disposition'] = f'attachment; filename={filename}'
            return res


# TODO: Switch this to a FormView
class AssemblyEditStateView(SingleAssemblyTeamMixin, View):
    def post(self, *args, **kwargs):
        request = self.request
        assembly = self.assembly

        value = request.POST.get('state', None)
        if value not in Assembly.State:
            messages.error(request, format_html('invalid state value "%s"', value))
            return redirect('backoffice:assemblyteam-detail', pk=assembly.pk)

        comment = request.POST.get('comment', '').strip()

        old_value = assembly.state
        assembly.state = value
        assembly.save(update_fields=['state'])
        assembly.log_activity(
            user=self.request.user,
            kind=ActivityLogEntry.Kind.STAFF,
            comment=comment if comment else None,
            state=ActivityLogChange(
                old=old_value,
                new=assembly.state,
            ),
        )

        # log the action
        messages.success(request, gettext('Assembly__edit__changed_state'))
        logger.info(
            'Assembly "%(assembly_slug)s" (%(assembly_pk)s): set state to "%(state)s" upon request by <%(user)s>.',
            {
                'assembly_slug': assembly.slug,
                'assembly_pk': assembly.pk,
                'state': value,
                'user': self.request.user.username,
            },
        )

        # invalidate export caches where necessary
        ConferenceExportCache.signal_assembly_modification(conference=self.conference, assembly=assembly)

        # we're done saving, redirect to edit view again
        return redirect('backoffice:assemblyteam-detail', pk=assembly.pk)

    def get(self, *args, **kwargs):
        new_value = self.request.GET.get('state')
        if new_value not in Assembly.State:
            messages.error(self.request, format_html('invalid state value "%s"', new_value))
            return redirect('backoffice:assemblyteam-detail', pk=self.assembly.pk)

        context = self.get_context_data()
        context['new_value'] = new_value
        new_state_label = next(label for k, label in Assembly.State.choices if k == new_value)
        context['new_value_translated'] = new_state_label
        return render(self.request, 'backoffice/assemblyteam_editstate.html', context)


class AssemblyEditHierarchyView(SingleAssemblyTeamMixin, View):
    def get_object(self, *args, **kwargs):
        assembly = self.assembly

        # say 'Not Found' if clusters aren't supported at all
        if not self.conference.support_clusters:
            raise Http404

        return assembly

    def post(self, *args, **kwargs):
        request = self.request
        assembly = self.get_object()

        old_value = assembly.hierarchy
        value = request.POST.get('value', None)
        if value not in Assembly.Hierarchy:
            messages.error(request, format_html('invalid hierarchy value "%s"', value))
            return redirect('backoffice:assemblyteam-detail', pk=assembly.pk)

        comment = request.POST.get('comment', '').strip()

        # don't allow changing cluster to regular if it has children
        if value == Assembly.Hierarchy.REGULAR and assembly.is_cluster and assembly.children.exists():
            messages.error(request, gettext('Assembly__edit__cluster_still_has_children'))
            return redirect(reverse('backoffice:assemblyteam-edithierarchy', kwargs={'pk': assembly.pk}) + '?value=' + value)

        assembly.hierarchy = value
        assembly.save(update_fields=['hierarchy'])

        # log the action
        messages.success(request, gettext('Assembly__edit__changed_hierarchy'))
        logger.info(
            'Assembly "%(assembly_slug)s" (%(assembly_pk)s): set hierarchy to "%(state)s" upon request by <%(user)s>.',
            {
                'assembly_slug': assembly.slug,
                'assembly_pk': assembly.pk,
                'state': value,
                'user': self.request.user.username,
            },
        )
        assembly.log_activity(
            user=self.request.user,
            kind=ActivityLogEntry.Kind.STAFF,
            comment=comment if comment else None,
            hierarchy=ActivityLogChange(
                old=old_value,
                new=assembly.hierarchy,
            ),
        )

        # invalidate export caches where necessary
        ConferenceExportCache.signal_assembly_modification(conference=self.conference, assembly=assembly)

        # we're done saving, redirect to edit view again
        return redirect('backoffice:assemblyteam-detail', pk=assembly.pk)

    def get(self, *args, **kwargs):
        new_value = self.request.GET.get('value')
        context = self.get_context_data()
        context['new_value'] = new_value
        context['new_value_translated'] = _(f'Assembly__hierarchy-{new_value}')  # noqa: INT001
        return render(self.request, 'backoffice/assemblyteam_edithierarchy.html', context)


class AssemblyEditPlacementView(SingleAssemblyTeamMixin, View):
    SRID = 0

    def post(self, *args, **kwargs):
        request = self.request
        assembly = self.assembly

        action = request.POST.get('action', 'save')
        inform_assembly = request.POST.get('inform_assembly', None) is not None
        poi = request.POST.get('location_point', None)
        boundaries = request.POST.get('location_boundaries', None)
        floor = request.POST.get('location_floor', None)
        comment = request.POST.get('comment', '').strip()

        old_values = {
            'state': assembly.state,
            'location_point': assembly.get_location_point_xy() or '',
            'location_boundaries': assembly.get_location_boundaries_xy() or '',
            'location_floor': str(assembly.get_location_floor_index() or ''),
        }
        changes = {}
        try:
            assembly.location_data = assembly.location_data or {}

            if poi and poi != '""':
                parsed = json.loads(poi)
                assert isinstance(parsed, list) and len(parsed) == 2

                # assembly.location_point = Point(x=parsed[0], y=parsed[1], srid=self.SRID)
                assembly.location_data['point'] = parsed
                changes['location_point'] = poi
            else:
                # assembly.location_point = None
                assembly.location_data['point'] = None
                changes['location_point'] = ''

            if boundaries and boundaries != '""':
                parsed = json.loads(boundaries)
                assert isinstance(parsed, list)

                # assembly.location_boundaries = MultiPolygon([Polygon(item) for item in parsed], srid=self.SRID)
                assembly.location_data['boundaries'] = parsed
                changes['location_boundaries'] = boundaries
            else:
                # assembly.location_boundaries = None
                assembly.location_data['boundaries'] = None
                changes['location_boundaries'] = ''

            if floor and floor != '""':
                parsed = int(floor)

                assembly.location_floor = self.conference.map_floors.filter(index=parsed).first()
                changes['location_floor'] = str(assembly.get_location_floor_index() or '')
            else:
                assembly.location_floor = None
                changes['location_floor'] = ''

        except ValueError:
            logger.exception('Failed to update position of assembly %s', assembly.pk)
            messages.error(request, gettext('Assembly__edit__position_error'))
            return redirect('backoffice:assemblyteam-editposition', pk=assembly.pk)

        if action == 'publish':
            # if assembly.location_point is not None or assembly.location_boundaries is not None:
            if assembly.location_data.get('point') or assembly.location_data.get('boundaries'):
                assembly.state = Assembly.State.PLACED
                changes['state'] = assembly.state
            else:
                messages.warning(request, gettext('Assembly__edit__position_missing_on_publish'))
                return redirect('backoffice:assemblyteam-editposition', pk=assembly.pk)

        elif action == 'recall':
            if assembly.is_placed:
                assembly.state = Assembly.State.ACCEPTED
                changes['state'] = assembly.state
            else:
                messages.warning(request, gettext('Assembly__edit__position_was_not_published'))

        elif action != 'save':
            messages.warning(request, gettext('Assembly__edit__position_unknown_action'))

        assembly.save(update_fields=['state', 'location_point', 'location_boundaries', 'location_floor', 'location_data'])

        # log the action
        messages.success(request, gettext('Assembly__edit__changed_position'))
        logger.info(
            'Assembly "%(assembly_slug)s" (%(assembly_pk)s): set POI to "%(poi)s" & boundaries to "%(boundaries)s" & floor to "%(floor)s" upon request by <%(user)s>.',  # noqa:E501
            {
                'assembly_slug': assembly.slug,
                'assembly_pk': assembly.pk,
                'floor': floor,
                'poi': poi,
                'boundaries': boundaries,
                'user': self.request.user.username,
            },
        )
        assembly.log_activity(
            user=self.request.user,
            kind=ActivityLogEntry.Kind.STAFF,
            comment=comment if comment else None,
            **{k: ActivityLogChange(old=old_values[k], new=v) for k, v in changes.items() if old_values[k] != v},
        )

        if assembly.state != old_values['state']:
            messages.success(request, gettext('Assembly__edit__changed_state'))
            logger.info(
                'Assembly "%(assembly_slug)s" (%(assembly_pk)s): changed state to "%(state)s" upon request by <%(user)s>.',
                {
                    'assembly_slug': assembly.slug,
                    'assembly_pk': assembly.pk,
                    'state': assembly.state,
                    'user': self.request.user.username,
                },
            )

        # invalidate export caches where necessary
        ConferenceExportCache.signal_assembly_modification(conference=self.conference, assembly=assembly)

        # inform assembly, if requested
        if inform_assembly:
            messages.warning(request, 'TODO: inform assembly')

        # we're done saving, redirect to edit view again
        return redirect('backoffice:assemblyteam-editposition', pk=assembly.pk)

    def get(self, *args, **kwargs):
        context = self.get_context_data()
        context['uses_map'] = True
        loc_data = context['assembly'].location_data or {}
        context['map_location'] = loc_data.get('point')  # context['assembly'].get_location_point_as_json()
        context['map_boundaries'] = loc_data.get('boundaries')  # context['assembly'].get_location_boundaries_as_json()
        return render(self.request, 'backoffice/assemblyteam_editposition.html', context)


class AssemblyMessageView(SingleAssemblyTeamMixin, FormView):
    template_name = 'backoffice/assemblyteam_message.html'
    form_class = AssemblyTeamMessageForm

    preview: bool = False

    def form_valid(self, form):
        action = self.request.POST.get('action', 'preview')
        assembly = self.assembly
        subject = form.cleaned_data['subject']
        message = form.cleaned_data['message']
        include_assembly = form.cleaned_data['include_assembly_name_in_subject']

        if action == 'send':
            # preview confirmed, send it
            try:
                recipients = assembly.send_mail_to_managers(subject=subject, message=message)
                assembly.log_activity(
                    user=self.request.user,
                    kind=ActivityLogEntry.Kind.MESSAGE,
                    subject=subject,
                    message=message,
                    recipients='; '.join(recipients),
                )
                messages.success(self.request, 'MESSAGE SENT')
                return redirect('backoffice:assemblyteam-detail', pk=assembly.pk)

            except Exception as err:
                messages.error(self.request, f'MESSAGE FAILED: {err}')

        # generate preview of mail
        p_subj, p_html, p_text = assembly.prepare_mail_to_managers(subject=subject, message=message, include_assembly_name=include_assembly)
        # render page w/ preview
        return render(
            self.request,
            self.template_name,
            {
                **self.get_context_data(),
                'preview': {
                    'subject': p_subj,
                    'html': p_html,
                    'text': p_text,
                },
            },
        )

    def get_context_data(self, *args, **kwargs):
        return {
            **super().get_context_data(*args, **kwargs),
            'recipients': '; '.join(
                list(
                    self.assembly.members.filter(
                        can_manage_assembly=True,
                        member__communication_channels__is_verified=True,
                        member__communication_channels__channel=UserCommunicationChannel.Channel.MAIL,
                    ).values_list('member__communication_channels__address', flat=True)
                )
            ),
        }
