import html
import re

import bleach
import mistletoe
from mistletoe.block_token import BlockToken
from mistletoe.html_renderer import HTMLRenderer
from mistletoe.span_token import AutoLink, Link, SpanToken, tokenize_inner

from django.db.models import Model
from django.urls import NoReverseMatch, reverse
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from modeltranslation.fields import build_localized_fieldname
from modeltranslation.settings import AVAILABLE_LANGUAGES

from .models import conference
from .utils import resolve_link


def markdown_header_slugify(value: str, separator: str) -> str:
    return 'md-' + slugify(value)


class PageLink(SpanToken):
    pattern = re.compile(r'\[\[ *([^|\]]+?) *(?:\| *(.*))? *\]\]')
    parse_group = 2
    repr_attributes = ('target', 'children', *SpanToken.repr_attributes)

    target: str

    def __init__(self, match):
        super().__init__(match)
        self.target = match.group(1)


class ProfileLink(SpanToken):
    pattern = re.compile(r'(?<!\S)@([-a-zA-Z0-9_]+)')
    parse_inner = False

    def __init__(self, match):
        self.username = match.group(1)


class Tag(SpanToken):
    pattern = re.compile(r'(?<!\S)#([-a-zA-Z0-9_]+)')
    parse_inner = False

    def __init__(self, match):
        self.tag_name = match.group(1)


class AlertBlock(BlockToken):
    pattern = re.compile(r'^!!! (danger|warning|info)(\s+".+")?\s*$')

    def __init__(self, match):
        self.alert_type = match[0]
        self.alert_caption = match[1].strip()[1:-1] if match[1] else None

        super().__init__(''.join(match[2]), tokenize_inner)

    @classmethod
    def start(cls, line):
        return cls.pattern.match(line)

    @classmethod
    def read(cls, lines):
        first_line = cls.pattern.match(next(lines))
        line_buffer = []
        while (next_line := lines.peek()) is not None and next_line.startswith('  '):
            line_buffer.append(next(lines))

        return first_line.group(1), first_line.group(2), line_buffer


class MyHtmlRenderer(HTMLRenderer):
    conf: 'conference.Conference'
    result: 'RenderResult'

    @staticmethod
    def render_html_block(token):
        return html.escape(token.content)

    @staticmethod
    def render_html_span(token):
        if token.content.startswith('<!--'):
            # HTML comments can be rendered as-is
            return token.content
        return html.escape(token.content)

    def render_alert_block(self, token: AlertBlock) -> str:
        result = f'<div class="mx-1 my-3 alert alert-{token.alert_type}">'
        if token.alert_caption:
            result += f'<p class="fw-bold">{token.alert_caption}</p>'
        result += self.render_inner(token)
        result += '</div>\n'
        return result

    def __init__(self, conf: 'conference.Conference', result: 'RenderResult', *extras, use_derefer_allowlist: bool = True, **kwargs):
        self.conf = conf
        self.result = result
        self.use_derefer_allowlist = use_derefer_allowlist
        super().__init__(PageLink, ProfileLink, Tag, AlertBlock, *extras, **kwargs)

    def render_link(self, token: Link) -> str:
        if token.target.startswith(('javascript:', 'data:')):
            token.target = ''

        link_type, url = resolve_link(token.target, self.use_derefer_allowlist)
        self.result.linked_urls.add(url)

        template = '<a href="{target}"{title} class="{link_type}">{inner}</a>'
        target = self.escape_url(url)
        if token.title:
            title = f' title="{html.escape(token.title)}"'
        else:
            title = ''
        inner = self.render_inner(token)
        return template.format(target=target, title=title, inner=inner, link_type=link_type)

    def render_auto_link(self, token: AutoLink) -> str:
        link_type, url = self.handle_link(token.target)
        self.result.linked_urls.add(url)

        template = '<a href="{target}" class="{link_type}">{inner}</a>'
        if token.mailto:
            target = f'mailto:{url}'
        else:
            target = self.escape_url(url)
        inner = self.render_inner(token)
        return template.format(target=target, inner=inner, link_type=link_type)

    def render_page_link(self, token: PageLink) -> str:
        from .models.pages import StaticPage
        # TODO namespaces für assemblies, rooms, tags, users, ...?

        self.result.linked_page_slugs.add(token.target)
        existing_page = StaticPage.objects.filter(conference=self.conf, slug=token.target).first()
        link_class = ' class="internal"' if existing_page else ' class="internal-nonexist"'

        if token.children:
            inner = self.render_inner(token)
        elif existing_page:
            inner = html.escape(existing_page.title)
        else:
            inner = ''

        if not inner:
            inner = html.escape(token.target)

        try:
            if existing_page:
                target = self.escape_url(reverse('plainui:static_page', kwargs={'page_slug': token.target}))
            else:
                target = self.escape_url(reverse('plainui:static_page_edit', kwargs={'page_slug': token.target}))
        except NoReverseMatch:
            target = ''

        template = '<a href="{target}"{link_class}>{inner}</a>'
        return template.format(target=target, link_class=link_class, inner=inner)

    def render_profile_link(self, token: ProfileLink) -> str:
        link_class = ' class="internal"'
        self.result.linked_user_slugs.add(token.username)
        try:
            target = self.escape_url(reverse('plainui:user', kwargs={'user_slug': token.username}))
        except NoReverseMatch:
            target = ''

        template = '<a href="{target}"{link_class}>@{username}</a>'
        return template.format(target=target, link_class=link_class, username=token.username)

    def render_tag(self, token: Tag) -> str:
        link_class = ' class="internal"'
        self.result.applied_tags.add(token.tag_name)
        try:
            target = self.escape_url(reverse('plainui:tag', kwargs={'tag_slug': token.tag_name}))
        except NoReverseMatch:
            target = ''

        template = '<a href="{target}"{link_class}>#{tag_name}</a>'
        return template.format(target=target, link_class=link_class, tag_name=token.tag_name)


class MySanitizedHtmlRenderer(MyHtmlRenderer):
    @staticmethod
    def render_html_block(token):
        if token.content.startswith('<!--'):
            # HTML comments can be rendered as-is
            return token.content
        return html.escape(token.content)

    @staticmethod
    def render_html_span(token):
        return html.escape(token.content)


class RenderResult:
    document: str
    linked_user_slugs: set[str]
    linked_page_slugs: set[str]
    applied_tags: set[str]
    linked_urls: set[str]

    def __init__(self):
        self.document = ''
        self.linked_user_slugs = set()
        self.linked_page_slugs = set()
        self.applied_tags = set()
        self.linked_urls = set()

    def merge_result(self, other_result: 'RenderResult'):
        self.linked_user_slugs |= other_result.linked_user_slugs
        self.linked_page_slugs |= other_result.linked_page_slugs
        self.applied_tags |= other_result.applied_tags
        self.linked_urls |= other_result.linked_urls


def render_markdown_ex(
    conf: 'conference.Conference', markup: str, allow_embedded_html: bool = False, sanitize_html: bool = True, dont_derefer_allowlist: bool = False
) -> RenderResult:
    # normalize newlines
    markup = markup.replace('\r\n', '\n').replace('\r', '\n')

    # remove HTML tags embedded in markdown, unless requested otherwise
    renderer = MyHtmlRenderer if allow_embedded_html else MySanitizedHtmlRenderer

    result = RenderResult()
    with renderer(conf, result) as renderer:
        renderer.use_derefer_allowlist = dont_derefer_allowlist
        rendered_markup = renderer.render(mistletoe.Document(markup))

    if sanitize_html:
        cleaner = bleach.Cleaner(
            tags=[
                *[
                    'br',
                    'h1',
                    'h2',
                    'h3',
                    'h4',
                    'h5',
                    'h6',
                    'ul',
                    'ol',
                    'li',
                    'dl',
                    'table',
                    'thead',
                    'tbody',
                    'th',
                    'td',
                    'tr',
                    'hr',
                    'p',
                    'pre',
                    'img',
                    'code',
                    'div',
                    'span',
                ],
                *list(bleach.sanitizer.ALLOWED_TAGS),
            ],
            attributes={
                '*': ['class', 'id'],
                'img': ['src', 'alt'],
                **bleach.sanitizer.ALLOWED_ATTRIBUTES,
            },
            protocols=['http', 'https', 'mailto', 'ftp', 'ftps'],
        )
        rendered_markup = cleaner.clean(rendered_markup)

    result.document = mark_safe(rendered_markup)
    return result


def render_markdown(
    conf: 'conference.Conference', markup: str, allow_embedded_html: bool = False, sanitize_html: bool = True, dont_derefer_allowlist: bool = False
) -> str:
    render_result = render_markdown_ex(
        conf, markup, allow_embedded_html=allow_embedded_html, sanitize_html=sanitize_html, dont_derefer_allowlist=dont_derefer_allowlist
    )
    return render_result.document


def store_relationships(conf, link_source, render_result: RenderResult):
    from django.contrib.contenttypes.models import ContentType

    from .models import MarkdownMeta
    from .models.markdown import MetaTypes

    src_type = ContentType.objects.get_for_model(link_source)
    src_object_id = link_source.pk

    MarkdownMeta.objects.filter(src_type=src_type, src_object_id=src_object_id).exclude(
        dst_type=MetaTypes.USER_SLUG, dst_object__in=render_result.linked_user_slugs
    ).exclude(dst_type=MetaTypes.PAGE_SLUG, dst_object__in=render_result.linked_page_slugs)

    MarkdownMeta.objects.bulk_create(
        [
            MarkdownMeta(src_type=src_type, src_object_id=src_object_id, dst_type=MetaTypes.USER_SLUG, dst_object=linked_user_slug)
            for linked_user_slug in render_result.linked_user_slugs
        ]
        + [
            MarkdownMeta(src_type=src_type, src_object_id=src_object_id, dst_type=MetaTypes.PAGE_SLUG, dst_object=linked_page_slug)
            for linked_page_slug in render_result.linked_page_slugs
        ],
        ignore_conflicts=True,
    )


def refresh_linking_markdown(conf: 'conference.Conference', link_target):
    from django.contrib.contenttypes.models import ContentType

    from .models import Assembly, BulletinBoardEntry, ConferenceMember, Event, MarkdownMeta, Room, StaticPage
    from .models.markdown import MetaTypes

    if isinstance(link_target, StaticPage):
        dst_type = MetaTypes.PAGE_SLUG
        dst_object = link_target.slug
    elif isinstance(link_target, ConferenceMember):
        dst_type = MetaTypes.USER_SLUG
        dst_object = link_target.user.slug
    else:
        raise Exception(f'Unsupported link_target type {type(link_target)!r}')

    relateds_by_type = {}
    for src_type_id, src_object_id in MarkdownMeta.objects.filter(dst_type=dst_type, dst_object=dst_object).values_list('src_type_id', 'src_object_id'):
        relateds_by_type.setdefault(src_type_id, []).append(src_object_id)

    for static_page in StaticPage.objects.filter(conference=conf, pk__in=relateds_by_type.get(ContentType.objects.get_for_model(StaticPage).pk, [])):
        static_page.body_html = render_markdown(conf, static_page.body)
        static_page.save()

    for conference_member in ConferenceMember.objects.filter(
        conference=conf, uuid__in=relateds_by_type.get(ContentType.objects.get_for_model(ConferenceMember).pk, [])
    ):  # noqa: E501
        conference_member.save(update_fields=['description'])

    for event in Event.objects.filter(conference=conf, pk__in=relateds_by_type.get(ContentType.objects.get_for_model(Event).pk, [])):
        event.save(update_fields=['description'])

    for room in Room.objects.filter(conference=conf, pk__in=relateds_by_type.get(ContentType.objects.get_for_model(Room).pk, [])):
        room.save(update_fields=['description'])

    for assembly in Assembly.objects.filter(conference=conf, pk__in=relateds_by_type.get(ContentType.objects.get_for_model(Assembly).pk, [])):
        assembly.save(update_fields=['description'])

    for assembly in BulletinBoardEntry.objects.filter(
        conference=conf, pk__in=relateds_by_type.get(ContentType.objects.get_for_model(BulletinBoardEntry).pk, [])
    ):  # noqa: E501
        assembly.save(update_fields=['description'])


def compile_translated_markdown_fields(
    obj: Model, conf: 'conference.Conference', field_name: str, dst_obj: Model | None = None, dst_field_name: str | None = None
) -> RenderResult:
    if dst_obj is None:
        dst_obj = obj
    if dst_field_name is None:
        dst_field_name = field_name + '_html'

    merged_results = RenderResult()
    for lang in AVAILABLE_LANGUAGES:
        localized_field_name = build_localized_fieldname(field_name, lang)
        localized_html_name = build_localized_fieldname(dst_field_name, lang)
        val = getattr(obj, localized_field_name)
        if val is None:
            setattr(dst_obj, localized_html_name, None)
        else:
            render_result = render_markdown_ex(conf, val)
            setattr(dst_obj, localized_html_name, render_result.document)
            merged_results.merge_result(render_result)

    return merged_results
