"""
Django settings for hub project.

Generated by 'django-admin startproject' using Django 3.1.1.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

import json
import os
import re
from contextlib import suppress
from logging import DEBUG
from pathlib import Path

import environ
from rich.console import Console
from rich.theme import Theme

from django.utils.translation import gettext_lazy as _

from hub.logging_utils import add_logging_level

# Ignore if the logging level already exists
# This may be the case when we are running this through gunicorn
with suppress(AttributeError):
    add_logging_level('TRACE', DEBUG - 5)


SCRIPT_NAME = os.getenv('SCRIPT_NAME', '')

env = environ.FileAwareEnv(
    # set casting, default value
    DJANGO_ENV_FILE=(str, ''),
    DJANGO_DEBUG=(str, ''),
    LOG_LEVEL=(str, 'WARN'),
    RICH_LOGGING=(bool, False),
    ALLOWED_HOSTS=(list, ['127.0.0.1', 'localhost']),
    MAIL_FROM=(str, 'webmaster+mail_from_not_configured@localhost'),
    ASSEMBLY_TEAM_FROM_EMAIL=(str, 'assembly_team+mail_from_not_configured@localhost'),
    ABUSE_USE_REPORTER_AS_FROM=(bool, True),
    ABUSE_DEFAULT_RECIPIENT=(str, 'report@cccv.de'),
    ABUSE_RECIPIENTS=(dict, {}),
    STATIC_URL=(str, f'{SCRIPT_NAME}/static/'),
    STORAGE_TYPE=(str, ''),
    MEDIA_URL=(str, f'{SCRIPT_NAME}/media/'),
    REDIS_URL=(str, None),
    COOKIE_NAME=(str, 'HUB_SESSION'),
    COOKIE_PATH=(str, SCRIPT_NAME),  # defaults to empty string, '/' will be used below in this case anyhow
    CLIENT_IP_HEADER=(str, None),
    DISABLE_RATELIMIT=(bool, False),
    BADGE_RATE_LIMIT=(str, '180/h'),
    SENTRY_ENDPOINT=(str, None),
    SENTRY_ENVIRONMENT=(str, 'production'),
    SENTRY_TRACES_SAMPLE_RATE=(float, 0.05),  # create a trace for this percentage of all requests
    SENTRY_PROFILES_SAMPLE_RATE=(float, 0.50),  # do a profiling for this percentage of _traced_ requests
    SSO_SECRET=(str, None),
    SSO_SECRET_GENERATE=(bool, False),
    PRETIX_ISSUER=(str, 'tickets.events.ccc.de'),
    PRETIX_SECRET=(str, None),
    METRICS_SERVER_IPS=(list, ['*']),
    TIMEZONE=(str, 'Europe/Berlin'),
    ARCHIVE_MODE=(bool, False),
    BIGBLUEBUTTON=(bool, False),
    BIGBLUEBUTTON_API_URL=(str, None),
    BIGBLUEBUTTON_API_TOKEN=(str, None),
    BIGBLUEBUTTON_END_MEETING_CALLBACK=(str, None),
    BIGBLUEBUTTON_INITIAL_PRESENTATION_URL=(str, None),
    HANGAR=(bool, False),
    HANGAR_URL=(str, None),
    ADMIN_BASE_URL=(str, '/c3admin'),
    API_BASE_URL=(str, '/api'),
    BACKOFFICE_BASE_URL=(str, '/backoffice'),
    PLAINUI_BASE_URL=(str, ''),
    SELECTED_CONFERENCE_ID=(str, None),
    PLAINUI_DEREFERER_URL=(str, '//dereferrer/{quoted_target}'),
    PLAINUI_DEREFERER_ALLOWLIST=(list, []),
    PLAINUI_DEREFERER_COUNTLIST=(list, []),
    PLAINUI_THEME_SUPPORT=(bool, False),
    ASSEMBLY_BANNER_FILE_SIZE_LIMIT=(int, 2),
    ASSEMBLY_BANNER_WIDTH_MINIMUM=(int, 100),
    ASSEMBLY_BANNER_HEIGHT_MINIMUM=(int, 100),
    ASSEMBLY_BANNER_WIDTH_MAXIMUM=(int, 2048),
    ASSEMBLY_BANNER_HEIGHT_MAXIMUM=(int, 1024),
    BADGE_IMAGE_WIDTH_MINIMUM=(int, 256),
    BADGE_IMAGE_HEIGHT_MINIMUM=(int, 256),
    BADGE_IMAGE_WIDTH_MAXIMUM=(int, 512),
    BADGE_IMAGE_HEIGHT_MAXIMUM=(int, 512),
    INVITATION_REJECTION_DEFAULT_TIMEOUT=(int, 3),
    EVENT_BANNER_WIDTH_MINIMUM=(int, 100),
    EVENT_BANNER_HEIGHT_MINIMUM=(int, 100),
    EVENT_BANNER_WIDTH_MAXIMUM=(int, 2048),
    EVENT_BANNER_HEIGHT_MAXIMUM=(int, 1024),
    USER_AVATAR_WIDTH_MINIMUM=(int, 64),
    USER_AVATAR_HEIGHT_MINIMUM=(int, 64),
    USER_AVATAR_WIDTH_MAXIMUM=(int, 1024),
    USER_AVATAR_HEIGHT_MAXIMUM=(int, 1024),
    API_USERS=(list, []),
    DISABLE_REQUEST_LOGGING=(bool, False),
    MOLLY_GUARD=(bool, True),
    CSP_DEFAULT_SRC=(list, ["'self'"]),
    CSP_SCRIPT_SRC=(list, ["'self'"]),
    CSP_STYLE_SRC=(list, ["'self'", "'unsafe-inline'"]),
    CSP_IMG_SRC=(list, ["'self'", 'data:']),
    CSP_CONNECT_SRC=(list, ["'self'"]),
    CSP_FONT_SRC=(list, ["'self'"]),
    CSP_OBJECT_SRC=(list, ["'none'"]),
    CSP_FRAME_SRC=(list, ["'none'"]),
    CSP_MEDIA_SRC=(list, ["'self'"]),
    CSP_FRAME_ANCESTORS=(list, ["'none'"]),
    CSP_FORM_ACTION=(list, ["'self'"]),
    CSP_BASE_URI=(list, ["'self'"]),
    CSP_INCLUDE_NONCE_IN=(list, ['script-src']),
    ADDITIONAL_LINK_PROTOCOLS=(dict, {}),
)


def gen_secret_key():
    """generates a value suitable for SECRET_KEY"""
    import random

    SECRET_CHARSET = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'
    return ''.join([random.SystemRandom().choice(SECRET_CHARSET) for i in range(50)])


# regex to match arbitrary hostnames
REGEX_HOST = re.compile(r'([\w\d-]+\.)+\w+(:\d+)?')


# helper function for parsing DEREFERER_*LIST (allow hostname only, regex and https:// prefixed strings)
def _handle_hostpattern_list(target: list, items: list):
    for item in items:
        if item.startswith('/') and item.endswith('/'):
            try:
                target.append(re.compile('https://' + item[1:-1]))
            except re.error as err:
                print('WARNING: failed to parse regex pattern "', item, "', ignoring it: ", err, sep='')
        elif item.startswith(('https://', 'http://')):
            target.append(item)
        elif REGEX_HOST.fullmatch(item):
            target.append('https://' + item)
        else:
            print('WARNING: failed to parse pattern "', item, "' (expected hostname, url or slash-enclosed regex), skipping it", sep='')


# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parents[2]

if env('DJANGO_ENV_FILE'):
    DJANGO_ENV_FILE_PATH = Path(env('DJANGO_ENV_FILE'))
    assert DJANGO_ENV_FILE_PATH.exists(), f'DJANGO_ENV_FILE path {DJANGO_ENV_FILE_PATH} does not exist!'
    environ.Env.read_env(DJANGO_ENV_FILE_PATH)

# prepare SECRET_KEY but must be overridden, i.e. in default.py
# by using using the local file approach for dev setups
SECRET_KEY = None

# don't run with debug turned on in production by default
DEBUG = env('DJANGO_DEBUG') == 'I_KNOW_WHAT_I_AM_DOING'

# allowed hosts may be multiple
ALLOWED_HOSTS = env('ALLOWED_HOSTS')

# read application version from marker file
with BASE_DIR.joinpath('version.json').open(encoding='UTF-8') as version_file:
    APP_VERSION_INFO = json.load(version_file)

# Application definition

INSTALLED_APPS = [
    # Django integrated stuff
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.humanize',
    'django.contrib.postgres',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.gis',
    # 3rd party extensions
    'modeltranslation',
    'oauth2_provider',
    'rest_framework',
    'rest_framework.authtoken',
    'django_rich',
    'csp',
    'rules',
    # our apps
    'core',
    'plainui',
]

MIDDLEWARE = [
    'core.middleware.SetRemoteAddrMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'oauth2_provider.middleware.OAuth2TokenMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'csp.middleware.CSPMiddleware',
    'core.middleware.TimezoneMiddleware',
    # TODO drüber nachdenken ob wir die brauchen (ist default an in Django)
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',  # noqa: ERA001
]

ROOT_URLCONF = 'hub.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.i18n',
                'csp.context_processors.nonce',
            ],
        },
    },
]

WSGI_APPLICATION = 'hub.wsgi.application'


# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.contrib.gis.db.backends.postgis',
        'NAME': 'hub',
    }
}


# Custom User model
AUTH_USER_MODEL = 'core.PlatformUser'


# default auto fields (legacy stuff for corsheaders app)
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'


# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

AUTHENTICATION_BACKENDS = (
    'rules.permissions.ObjectPermissionBackend',
    'oauth2_provider.backends.OAuth2Backend',
    'django.contrib.auth.backends.ModelBackend',
)

# Session Cookie configuration
SESSION_COOKIE_NAME = env('COOKIE_NAME')
SESSION_COOKIE_HTTPONLY = True  # session cookie is unavailable to JavaScript (default)
SESSION_COOKIE_SAMESITE = 'Lax'  # set SameSite=Lax (default)
SESSION_COOKIE_PATH = env('COOKIE_PATH') or '/'  # use configured path, SESSION_NAME or default '/'
SESSION_COOKIE_SECURE = True  # mark session cookie as https-only
SESSION_SAVE_EVERY_REQUEST = False  # no need to update a session on each request (default)

# CSRF Cookie configuration
CSRF_COOKIE_NAME = env('CSRF_COOKIE_NAME', default=SESSION_COOKIE_NAME.replace('_SESSION', '_CSRF') if '_SESSION' in SESSION_COOKIE_NAME else 'HUB_CSRF')
CSRF_COOKIE_PATH = SESSION_COOKIE_PATH
CSRF_COOKIE_SECURE = SESSION_COOKIE_SECURE

# Content Security Policy
CSP_DEFAULT_SRC = env('CSP_DEFAULT_SRC')
CSP_SCRIPT_SRC = env('CSP_SCRIPT_SRC')
CSP_STYLE_SRC = env('CSP_STYLE_SRC')
CSP_IMG_SRC = env('CSP_IMG_SRC')
CSP_CONNECT_SRC = env('CSP_CONNECT_SRC')
CSP_FONT_SRC = env('CSP_FONT_SRC')
CSP_OBJECT_SRC = env('CSP_OBJECT_SRC')
CSP_FRAME_SRC = env('CSP_FRAME_SRC')
CSP_MEDIA_SRC = env('CSP_MEDIA_SRC')
CSP_FRAME_ANCESTORS = env('CSP_FRAME_ANCESTORS')
CSP_FORM_ACTION = env('CSP_FORM_ACTION')
CSP_BASE_URI = env('CSP_BASE_URI')
CSP_INCLUDE_NONCE_IN = env('CSP_INCLUDE_NONCE_IN')

# OAuth2 configuration
OAUTH2_PROVIDER_APPLICATION_MODEL = 'core.Application'
OAUTH2_PROVIDER = {
    'ALLOWED_REDIRECT_URI_SCHEMES': ['https'],
    'SCOPES_BACKEND_CLASS': 'core.sso.OAuth2Scopes',
}

# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = 'en'
LANGUAGE_COOKIE_NAME = env(
    'LANGUAGE_COOKIE_NAME', default=SESSION_COOKIE_NAME.replace('_SESSION', '_LANG') if '_SESSION' in SESSION_COOKIE_NAME else 'HUB_LANG'
)  # noqa:E501
LANGUAGE_COOKIE_PATH = SESSION_COOKIE_PATH
LANGUAGE_COOKIE_SECURE = SESSION_COOKIE_SECURE

TIME_ZONE = env('TIMEZONE')
USE_I18N = True
USE_TZ = True
LANGUAGES = [
    ('de', _('German')),
    ('en', _('English')),
]

MODELTRANSLATION_FALLBACK_LANGUAGES = ['en', 'de']

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

# If True a route to serve the collected static files will be added
SERVE_STATIC = False

STATIC_URL = env('STATIC_URL')
STATIC_ROOT = BASE_DIR / 'static.dist'

# setup storage, we assume the default
STORAGES = {
    'staticfiles': {
        'BACKEND': 'django.contrib.staticfiles.storage.ManifestStaticFilesStorage',
    },
}

if env.bool('USE_PLAIN_STATICFILES', False):
    STORAGES['staticfiles']['BACKEND'] = 'django.contrib.staticfiles.storage.StaticFilesStorage'

storage_type = env('STORAGE_TYPE').lower()
if storage_type == 's3':
    STORAGES['default'] = {
        'BACKEND': 'storages.backends.s3boto3.S3Boto3Storage',
        'OPTIONS': {
            'file_overwrite': False,
        },
    }

    AWS_STORAGE_BUCKET_NAME = env.str('S3_BUCKET', 'static')

    AWS_ACCESS_KEY_ID = env.str('S3_ACCESS_KEY')
    AWS_SECRET_ACCESS_KEY = env.str('S3_SECRET_KEY')
    AWS_S3_ENDPOINT_URL = env.str('S3_ENDPOINT')

    # allow a custom domain to serve the media files, the used path must _not_ end with a slash
    media_url = env.url('MEDIA_URL', default=AWS_S3_ENDPOINT_URL)
    AWS_S3_CUSTOM_DOMAIN = media_url.netloc
    AWS_LOCATION = media_url.path.strip('/') if media_url.path not in [None, '', '/'] else ''

elif storage_type == 'sftp':
    STORAGES['default'] = {'BACKEND': 'storages.backend.sftpstorage.SFTPStorage'}
    SFTP_STORAGE_HOST = env.str('SFTP_HOST')
    SFTP_STORAGE_ROOT = env.str('SFTP_DIR')
    SFTP_STORAGE_INTERACTIVE = False
    SFTP_STORAGE_PARAMS = env.dict('SFTP_PARAMS')

elif storage_type in ('local', ''):
    if storage_type == '':
        import warnings

        warnings.warn('No STORAGE_TYPE selected, defaulting to "local". Verify if that is what you want!', RuntimeWarning)

    STORAGES['default'] = {
        'BACKEND': 'django.core.files.storage.FileSystemStorage',
        'OPTIONS': {
            'base_url': env('MEDIA_URL'),
            'location': env.path('MEDIA_PATH', BASE_DIR / 'media'),
        },
    }

else:
    raise ValueError(f'Please specify storage type (s3/sftp/local), "{storage_type}" is unknown.')

# Media (uploaded files)
MEDIA_URL = env('MEDIA_URL')
MEDIA_ROOT = BASE_DIR / 'media'

# serve media files in MEDIA_ROOT at /media/ (ignoring configured MEDIA_URL, discouraged in production!)
FORCE_SERVE_MEDIA = False

# REST Framework Config
REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': [
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ]
}

# Cache
if env('REDIS_URL'):
    CACHES = {
        'default': {
            'BACKEND': 'django_redis.cache.RedisCache',
            'LOCATION': env('REDIS_URL'),
            'OPTIONS': {
                'CLIENT_CLASS': 'django_redis.client.DefaultClient',
                'SERIALIZER': 'django_redis.serializers.json.JSONSerializer',
            },
        }
    }
else:
    CACHES = {
        'default': {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        }
    }

# Header to look up real client IP in. Used for logging / rate limiting purposes. `None` = don't lookup, looked up in `request.META`
# see https://docs.djangoproject.com/en/3.2/ref/request-response/#django.http.HttpRequest.META for details
CLIENT_IP_HEADER = env('CLIENT_IP_HEADER')
RATELIMIT_ENABLE = not env('DISABLE_RATELIMIT')
BADGE_RATE_LIMIT = env('BADGE_RATE_LIMIT')

# List of disallowed assembly slugs
FORBIDDEN_ASSEMBLY_SLUGS = ['admin', 'visit', 'maps', 'api', 'pusher']

# Logging configuration
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'rich': {
            'datefmt': '[%Y-%m-%d %H:%M:%S %z]',
        },
    },
    'handlers': {
        'console': {
            'class': 'hub.logging_utils.HubLoggingHandler',
            'time_zone': TIME_ZONE,
            'formatter': 'rich',
            'console': Console(
                theme=Theme({'logging.level.trace': 'light_cyan1'}),
                no_color=(not env('RICH_LOGGING')),
            ),
            'show_path': env('RICH_LOGGING'),
            'markup': env('RICH_LOGGING'),
            'rich_tracebacks': env('RICH_LOGGING'),
            'tracebacks_show_locals': env('RICH_LOGGING'),
        },
        'null': {
            'class': 'logging.NullHandler',
        },
    },
    'loggers': {
        'django.request': {
            'handlers': ['console'] if not env('DISABLE_REQUEST_LOGGING') else ['null'],
            'propagate': False,
        },
    },
    'root': {
        'handlers': ['console'],
        'level': env('LOG_LEVEL'),
    },
}

# Mail configuration
DEFAULT_FROM_EMAIL = env.str('MAIL_FROM')
SERVER_EMAIL = DEFAULT_FROM_EMAIL
MAIL_REPLY_TO = []
MAIL_SUBJECT_PREFIX = env.str('MAIL_SUBJECT_PREFIX', default='')
MAIL_SIGNATURE = env.str('MAIL_SIGNATURE', default='')
SUPPORT_HTML_MAILS = True

# Abuse / Report mails
ABUSE_USE_REPORTER_AS_FROM = env.bool('ABUSE_USE_REPORTER_AS_FROM')
ABUSE_DEFAULT_RECIPIENT = env.str('ABUSE_DEFAULT_RECIPIENT')
ABUSE_RECIPIENTS = env.dict('ABUSE_RECIPIENTS')

# Assembly team mail configuration
ASSEMBLY_TEAM_FROM_EMAIL = env.str('ASSEMBLY_TEAM_FROM_EMAIL')
ASSEMBLY_TEAM_REPLY_TO = [env.str('ASSEMBLY_TEAM_FROM_EMAIL')]


# try to parse mail config as URL, e.g. smtp+tls://user%40gmail.com:password@mx.local:587
mail_config_str = env.str('MAIL_CONFIG', default=None)
if mail_config_str:
    mail_config = environ.FileAwareEnv.email_url_config(mail_config_str)
    # update local vars with the parsed EMAIL_HOST, EMAIL_HOST_USER, etc.
    vars().update(mail_config)

# Debugging Tools
HEADERS_DEBUG_ENDPOINT_TOKEN = None

# Sentry
SENTRY_ENDPOINT = env('SENTRY_ENDPOINT')
SENTRY_ENVIRONMENT = env('SENTRY_ENVIRONMENT')
SENTRY_TRACES_SAMPLE_RATE = env('SENTRY_TRACES_SAMPLE_RATE')
SENTRY_PROFILES_SAMPLE_RATE = env('SENTRY_PROFILES_SAMPLE_RATE')

# API access
API_USERS = env('API_USERS')

# SSO
SSO_COOKIE_NAME = 'SSO_TOKEN'
SSO_COOKIE_DOMAIN = '.localhost'
SSO_HEADER = 'X-SSO-TOKEN'

SSO_SECRET = env('SSO_SECRET')
if SSO_SECRET is None and env('SSO_SECRET_GENERATE'):
    SSO_SECRET = gen_secret_key()

# Pretix Token integration, see core.models.ticket.ConferenceMemberTicket.validate_pretix_ticket and plainui.views.RedeemToken
PRETIX_ISSUER = env('PRETIX_ISSUER')  # expected value in the 'iss' field of pretix
PRETIX_SECRET_KEY = env('PRETIX_SECRET')  # the JWT shared secret with Pretix

# ----------------------------------
# Metrics
# ----------------------------------

# Restrict the access to /metrics/ to certain IP addresses (e.g. allowe the Prometheus/Grafana server only).
# You can add one or multiple IP addresse by setting the `METRICS_SERVER_IPS` env var.
# e.g. METRICS_SERVER_IPS="127.0.0.1,80.147.140.51" # noqa: ERA001
# The default is to allow every IP address (METRICS_SERVER_IPS="*").
METRICS_SERVER_IPS = env('METRICS_SERVER_IPS')


# ----------------------------------
# additional protocols supported by core.utils.resolve_internal_url() and thus, all markdown in the hub
# ----------------------------------

ADDITIONAL_LINK_PROTOCOLS = env.dict('ADDITIONAL_LINK_PROTOCOLS', cast={'value': str})

# ----------------------------------
# External Schedule Support
# ----------------------------------

SCHEDULE_SUPPORT = [
    'core.schedules.schedulexml.ScheduleXMLSupport',
    'core.schedules.schedulejson.ScheduleJSONSupport',
    'core.schedules.schedulejsonfrab.ScheduleJSONFrabSupport',
]

# the file:// protocol should only be used in the unit tests and _not_ in production!
SCHEDULES_SUPPORT_FILE_PROTOCOL = False

# ----------------------------------
# External Components Integration
# ----------------------------------

# BigBlueButton
INTEGRATIONS_BBB = env('BIGBLUEBUTTON')
BIGBLUEBUTTON_API_URL = env('BIGBLUEBUTTON_API_URL')
BIGBLUEBUTTON_API_TOKEN = env('BIGBLUEBUTTON_API_TOKEN')
BIGBLUEBUTTON_END_MEETING_CALLBACK = env('BIGBLUEBUTTON_END_MEETING_CALLBACK')  # url bbb will call to notify us of ending meetings
BIGBLUEBUTTON_INITIAL_PRESENTATION_URL = env('BIGBLUEBUTTON_INITIAL_PRESENTATION_URL')  # url to tell bbb to use as initial presentation

# Hangar
INTEGRATIONS_HANGAR = env('HANGAR')
HANGAR_URL = env('HANGAR_URL')

# ----------------------------------
# Base URL Configurations
# ----------------------------------

# base domain of the elements (used to generate links to other APPS)

ADMIN_BASE_URL = env('ADMIN_BASE_URL')
API_BASE_URL = env('API_BASE_URL')
BACKOFFICE_BASE_URL = env('BACKOFFICE_BASE_URL')
PLAINUI_BASE_URL = env('PLAINUI_BASE_URL')

# Database ID of the active conference for frontend (PlainUI) and/or API.
# To initialize create a conference in the backend, then configure the conference, then restart the UI workers
SELECTED_CONFERENCE_ID = env('SELECTED_CONFERENCE_ID')

# ----------------------------------
# PlainUI Configuration
# ----------------------------------

# archive mode disables user login etc.
ARCHIVE_MODE = env('ARCHIVE_MODE')

# Absolute url to use when dereferring links. Replaces `{quoted_target}` with the urlquoted target url
# If you modify this on a non-empty database you should run `manage.py rerender_markdown` to update cashed markdown that contains old urls.
PLAINUI_DEREFERER_URL = env('PLAINUI_DEREFERER_URL')

# URI patterns that are auto-allowed (never get dereferrer'd)
# Only looks at schema and host/port, so for example for the url https://rc3.world:4321/some/where/ the entry should be `'https://rc3.world'`.
# Can also be a regex, for example re.compile(r'https://.*\.rc3\.world')
# DON'T ADD TRAILING SLASHES
DEREFERRER_GLOBAL_ALLOWLIST = []
_handle_hostpattern_list(DEREFERRER_GLOBAL_ALLOWLIST, env('PLAINUI_DEREFERER_ALLOWLIST'))

# URI patterns where following outgoing links is counted in the database
# rules are the same as with DEREFERRER_GLOBAL_ALLOWLIST
DEREFERRER_COUNT_ACCESS = []
_handle_hostpattern_list(DEREFERRER_COUNT_ACCESS, env('PLAINUI_DEREFERER_COUNTLIST'))

# flag if newly created pages are localized by default (or not)
STATIC_PAGE_LOCALIZED_BY_DEFAULT = env.bool('STATICPAGES_LOCALIZED_BY_DEFAULT', False)
# Time after which a non-refreshed static page lock will expire (in seconds)
STATIC_PAGE_LOCK_TIMEOUT = env.int('STATICPAGES_LOCK_TIMEOUT', 5 * 60)

PLAINUI_THEME_SUPPORT = env('PLAINUI_THEME_SUPPORT')

ASSEMBLY_BANNER_FILE_SIZE_LIMIT = env('ASSEMBLY_BANNER_FILE_SIZE_LIMIT')
ASSEMBLY_BANNER_WIDTH_MINIMUM = env('ASSEMBLY_BANNER_WIDTH_MINIMUM')
ASSEMBLY_BANNER_HEIGHT_MINIMUM = env('ASSEMBLY_BANNER_HEIGHT_MINIMUM')
ASSEMBLY_BANNER_WIDTH_MAXIMUM = env('ASSEMBLY_BANNER_WIDTH_MAXIMUM')
ASSEMBLY_BANNER_HEIGHT_MAXIMUM = env('ASSEMBLY_BANNER_HEIGHT_MAXIMUM')

BADGE_IMAGE_WIDTH_MINIMUM = env('BADGE_IMAGE_WIDTH_MINIMUM')
BADGE_IMAGE_HEIGHT_MINIMUM = env('BADGE_IMAGE_HEIGHT_MINIMUM')
BADGE_IMAGE_WIDTH_MAXIMUM = env('BADGE_IMAGE_WIDTH_MAXIMUM')
BADGE_IMAGE_HEIGHT_MAXIMUM = env('BADGE_IMAGE_HEIGHT_MAXIMUM')

EVENT_BANNER_WIDTH_MINIMUM = env('EVENT_BANNER_WIDTH_MINIMUM')
EVENT_BANNER_HEIGHT_MINIMUM = env('EVENT_BANNER_HEIGHT_MINIMUM')
EVENT_BANNER_WIDTH_MAXIMUM = env('EVENT_BANNER_WIDTH_MAXIMUM')
EVENT_BANNER_HEIGHT_MAXIMUM = env('EVENT_BANNER_HEIGHT_MAXIMUM')

USER_AVATAR_WIDTH_MINIMUM = env('USER_AVATAR_WIDTH_MINIMUM')
USER_AVATAR_HEIGHT_MINIMUM = env('USER_AVATAR_HEIGHT_MINIMUM')
USER_AVATAR_WIDTH_MAXIMUM = env('USER_AVATAR_WIDTH_MAXIMUM')
USER_AVATAR_HEIGHT_MAXIMUM = env('USER_AVATAR_HEIGHT_MAXIMUM')

# Set default timeout for when invitation is possible after rejection (in hours)
INVITATION_REJECTION_DEFAULT_TIMEOUT = env('INVITATION_REJECTION_DEFAULT_TIMEOUT')

MOLLY_GUARD = env('MOLLY_GUARD')

# Django 6 tranisitional settings, this can be removed once we are on Django 6,
# As this setting is also removed in Django 6, it will show a deprecation warning until then.
FORMS_URLFIELD_ASSUME_HTTPS = True
